/******************************************************************************
 *
 * brica/functor.hpp
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *
 *****************************************************************************/

#ifndef __BRICA_FUNCTOR_HPP__
#define __BRICA_FUNCTOR_HPP__

#include "brica/assocvec.hpp"
#include "brica/buffer.hpp"

#include <functional>
#include <string>

namespace brica {

using Dict = AssocVec<std::string, Buffer>;
using Functor = std::function<void(Dict&, Dict&)>;

}  // namespace brica

#endif  // __BRICA_FUNCTOR_HPP__
